/*!
 * @file grpc_server.cpp
 *
 * @brief This file contains the implementation of the GrpcSyncWorker class.
 *
 * Copyright 2023 (c) RDI Technologies as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * RDI Technologies. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with RDI Technologies.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>

#include <chrono>
#include <condition_variable>
#include <cstdint>
#include <memory>
#include <mutex>
#include <string>
#include <thread>
#include <utility>
#include <vector>

#include <fmt/format.h>

#include <grpcpp/grpcpp.h>
#include <grpcpp/security/server_credentials.h>
#include <grpcpp/server.h>
#include <grpcpp/server_builder.h>
#include <grpcpp/server_context.h>

#include <json/json.h>

#include <fastec/fi_ioctl.h>
#include <fastec/gige_msg.h>
#include <fastec/interval_timer.h>
#include <fastec/fas_log.h>
#include <fastec/fas_thread.h>
#include <fastec/fas_trace.h>

#include "grpc_server.h"

#include "base64.h"
#include "bounce.h"
#include "camera_server.h"
#include "cs_utils.h"
#include "fcp.h"
#include "fpga.h"
#include "image_io.h"
#include "ipm_proc.h"
#include "jpeg_proc.h"
#include "jpeg_util.h"
#include "part_mgr.h"
#include "pb_control.h"

#include "cam_server.grpc.pb.h"

using namespace RDICamServer;
using namespace grpc;

namespace GRPC_SERVER
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

enum GRPC_TRACE
{
    GRPC_FUNC       = 0x00000010,
    GRPC_READREG    = 0x00000020,
    GRPC_WRITEREG   = 0x00000040,
    GRPC_STREAM     = 0x00000080,
    GRPC_NOTIFY     = 0x00000100,
    GRPC_REQUESTOR  = 0x00000200,
    GRPC_VERBOSE    = 0x00002000,
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
    GRPC_FUNC       |
    GRPC_REQUESTOR  |
    // GRPC_READREG |
    // GRPC_WRITEREG |
    // GRPC_STREAM |
    // GRPC_NOTIFY |
    // GRPC_VERBOSE |
    // TRACE_0;        // Set our trace level to TRACE_0 to trace.
    TRACE_3;        // Set our trace level to TRACE_3 to not trace.

//--------------------------------------------------------------------
//! @brief The GrpcRequestor class.
//--------------------------------------------------------------------

//! Message types for the requestor class.
enum GREQ_MESSAGE_TYPES
{
    GREQ_CONTROL = 0x30000          //!< Send a message (read or write a register) to the camera server.
};

//! Message class for GREQ_CONTROL messages.
//! Note that this is a SYNC message (i.e. blocking).
class GreqMessage : public FAS_THREAD::SyncMessage
{
public:
    GreqMessage() : FAS_THREAD::SyncMessage()
    {
        req = (FAS_THREAD::REQ_CODE)GREQ_CONTROL;
        memset(&msg, 0, sizeof(msg));
    }

    virtual ~GreqMessage() = default;

    //! The message which is to be sent to and received from
    //! the camera server.
    CS_NETWORK::CS_NETWORK_MESSAGE msg;
};

class GrpcRequestor : public FAS_THREAD::MsgThread
{
public:
    GrpcRequestor();
    virtual ~GrpcRequestor() = default;

    void stop() override;

    static void receiveResponse(const CS_NETWORK::CS_NETWORK_MESSAGE *msgPtr);

protected:
    void messageHandle(FAS_THREAD::Message *message) override;

private:
    bool handleRequest(GreqMessage *msg);

    static std::mutex work_mutex;
    static std::condition_variable work_cond;
    static bool response_received;
    static CS_NETWORK::CS_NETWORK_MESSAGE response_msg;
};

std::mutex GrpcRequestor::work_mutex;
std::condition_variable GrpcRequestor::work_cond;
bool GrpcRequestor::response_received(false);
CS_NETWORK::CS_NETWORK_MESSAGE GrpcRequestor::response_msg;

GrpcRequestor::GrpcRequestor() : FAS_THREAD::MsgThread()
{
    TP(GRPC_REQUESTOR, "%s", __PRETTY_FUNCTION__);
}

void GrpcRequestor::stop()
{
    // Set the stop request flag.
    stop_request = true;

    // Make sure we wake up the thread.
    work_cond.notify_all();
}

/*!
 * @brief Process the messages for the thread.
 */
void GrpcRequestor::messageHandle(FAS_THREAD::Message *message)
{
    int req_code = (int)message->req;
    TP(GRPC_REQUESTOR, "%s: req=%d", __PRETTY_FUNCTION__, req_code);

    try
    {
        // Parse the message.
        switch (req_code)
        {
        case GREQ_CONTROL:
            if (handleRequest(static_cast<GreqMessage *>(message)))
                message->resp = FAS_THREAD::OK;
            else
                message->resp = FAS_THREAD::FAULT;
            break;

        case FAS_THREAD::DUMP:
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::HALT:
            // Remember that we've stopped.
            halted = true;

            // All done.
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_ERR, "Exception in GrpcRequestor:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
    }

    // Forward on the messsage.
    message->forward();

    TP(GRPC_REQUESTOR, "%s: req=%d done", __PRETTY_FUNCTION__, req_code);
}

bool GrpcRequestor::handleRequest(GreqMessage *msg)
{
    bool rc(false);

    // Sanity check.
    if (msg == nullptr)
        return false;

    // Are we being asked to stop?
    if (stop_request || halted)
        return false;

    TP(GRPC_REQUESTOR, "%s: req_id=%d", __PRETTY_FUNCTION__, htons(msg->msg.message.header.req.req_id));

    // Create a block, so we can control access to our response buffer.
    std::unique_lock<std::mutex> locker(work_mutex);

    // Reset our response flag and message.
    response_received = false;
    memset(&response_msg, 0, sizeof(response_msg));

    // Pass on the request.
    CameraServer::add_work(&msg->msg);

    // Do we expect a response?
    if ((msg->msg.message.header.req.flag & GIGE_FLAG_MUST_ACK_BE) == GIGE_FLAG_MUST_ACK_BE)
    {
        // Wait for the response.
        int retries;
        const int MAX_RETRY_COUNT(1000);
        const int SLEEP_MS(50);
        for (retries = 0; retries < MAX_RETRY_COUNT; ++retries)
        {
            // Did we get a response?
            if (response_received)
            {
                if (response_msg.message.header.req.req_id == msg->msg.message.header.req.req_id)
                {
                    TP(GRPC_REQUESTOR, "%s: Got response for ID %d [retries=%d]", __PRETTY_FUNCTION__, htons(response_msg.message.header.req.req_id), retries);
                    break;
                }
                else
                {
                    fi_log(LOG_INFO, "%s: Received response for %d instead of %d!", __PRETTY_FUNCTION__, htons(response_msg.message.header.req.req_id), htons(msg->msg.message.header.req.req_id));
                }
            }
            else
            {
                // Delay for a bit before checking the queue and state again.
                std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
                using ms = std::chrono::milliseconds;
                work_cond.wait_until(locker, now + ms(SLEEP_MS));
                TP(GRPC_REQUESTOR, "%s: [%d] response_received=%d", __PRETTY_FUNCTION__, retries, response_received);
            }

            // Are we being asked to stop?
            if (stop_request)
                return false;
        }

        // Did we get a response?
        if (!response_received)
        {
            fi_log(LOG_ERR, "%s: No response received for ID %d", __PRETTY_FUNCTION__, htons(msg->msg.message.header.req.req_id));

            std::string output_string;
            uint16_t cmd = htons(msg->msg.message.header.req.command);
            switch (cmd)
            {
            case GIGE_MSG_READREG_CMD:
                output_string = fmt::format("cmd=0x{0:x}[GIGE_MSG_READREG_CMD]", cmd);
                break;
            case GIGE_MSG_READMEM_CMD:
                output_string = fmt::format("cmd=0x{0:x}[GIGE_MSG_READMEM_CMD]", cmd);
                break;
            case GIGE_MSG_WRITEREG_CMD:
                output_string = fmt::format("cmd=0x{0:x}[GIGE_MSG_WRITEREG_CMD]", cmd);
                break;
            default:
                output_string = fmt::format("cmd=0x{0:x}[unknown]", cmd);
                break;
            }
            uint32_t reg_addr = htonl(msg->msg.message.data_l[0]);
            output_string += fmt::format(", addr=0x{0:x}", reg_addr);
            fi_log(LOG_ERR, "%s: %s", __PRETTY_FUNCTION__, output_string.c_str());

            msg->resp = FAS_THREAD::FAULT;
        }
        else
        {
            // Copy over our response.
            memcpy(&msg->msg, &response_msg, sizeof(response_msg));
            rc = true;
        }
    }
    else
    {
        // No response expected, so return success.
        rc = true;
    }

    // All done.
    return rc;
}

void GrpcRequestor::receiveResponse(const CS_NETWORK::CS_NETWORK_MESSAGE *msgPtr)
{
    TP(GRPC_REQUESTOR, "%s: req_id=%d", __PRETTY_FUNCTION__, htons(msgPtr->message.header.req.req_id));

    // Protect access to the response message structure.
    {
        std::unique_lock<std::mutex> locker(work_mutex);

        // Copy over the  response information.
        memcpy(&response_msg, msgPtr, sizeof(CS_NETWORK::CS_NETWORK_MESSAGE));
        response_received = true;
    }

    // Wake up the worker.
    work_cond.notify_all();
}

//--------------------------------------------------------------------
//! @brief The GrpcSyncWorker class.
//--------------------------------------------------------------------

typedef struct IMAGE_STREAM_INFO
{
    bool enable;
    ::grpc::ServerReaderWriter<ImageData, StreamEnable> *stream;
} IMAGE_STREAM_INFO;

typedef struct NOTIFY_INFO
{
    bool enable;
    ::grpc::ServerReaderWriter<AsyncNotifyResponse, NotifyEnable> *stream;
} NOTIFY_INFO;

class GrpcSyncWorker final : public RDICamServer::RDICameraService::Service
{
public:
    GrpcSyncWorker();
    ~GrpcSyncWorker();

    ::grpc::Status AsyncNotify(::grpc::ServerContext *context, ::grpc::ServerReaderWriter<::RDICamServer::AsyncNotifyResponse, ::RDICamServer::NotifyEnable> *stream) override;
    ::grpc::Status ImageStream(::grpc::ServerContext *context, ::grpc::ServerReaderWriter<::RDICamServer::ImageData, ::RDICamServer::StreamEnable> *stream) override;
    ::grpc::Status ReadRegister(::grpc::ServerContext *context, const ::RDICamServer::ReadRegisterRequest *request, ::RDICamServer::JsonResponse *response) override;
    ::grpc::Status ReadMultipleRegister(::grpc::ServerContext *context, const ::RDICamServer::MultipleRegisterRequest *request, ::RDICamServer::JsonResponse *response) override;
    ::grpc::Status WriteRegisters(::grpc::ServerContext *context, const ::RDICamServer::WriteRegistersRequest *request, ::RDICamServer::JsonResponse *response) override;

    void close();
    bool sendNotification(uint32_t reg_address, uint32_t reg_value);
    bool streamIsReady();
    bool streamSendImage(const GRPC_IMAGE& image);

private:
    void buildReadRequest(uint32_t reg_addr, uint16_t& read_cmd, CS_NETWORK::CS_NETWORK_MESSAGE& msg);
    Json::Value buildReadResponse(uint32_t read_cmd, uint32_t reg_addr, fcp_msg *message_ptr);
    int regReadCmdType(uint32_t reg_addr, uint16_t& reg_length);

private:
    volatile bool closed;

    std::mutex notify_mutex;                //!< Control access to the notification stream pointer.
    std::vector<NOTIFY_INFO *> notify_stream;

    std::mutex stream_mutex;                //!< Control access to the stream pointer.
    std::vector<IMAGE_STREAM_INFO *> image_stream;

    uint16_t msg_id;                        //!< We need an incrementing ID for the FCP messages.

    std::unique_ptr<GrpcRequestor> requestor;
};

//--------------------------------------------------------------------
//! @brief The GrpcServer class.
//!
//! The GrpcServer class is a Singleton object, as there can be
//! only one.
//--------------------------------------------------------------------

class GrpcServer final : public FAS_THREAD::Thread
{
private:
    static std::unique_ptr<GrpcServer> instance;
    static std::mutex instance_mutex;

    static bool isClosed;

private:
    GrpcServer();

public:
    ~GrpcServer();

    // Singletons should not be cloneable.
    GrpcServer(GrpcServer& other) = delete;

    // Singletons should not be assignable.
    void operator=(const GrpcServer&) = delete;

    /**
     * This is the static method that controls the access to the GrpcServer
     * instance. On the first run, it creates a GrpcServer object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static GrpcServer *factory();

    static void close()
    {
        TP(GRPC_FUNC, "%s", __PRETTY_FUNCTION__);

        // Make sure we haven't already closed.
        if (!isClosed)
        {
            std::lock_guard<std::mutex> lock(instance_mutex);

            // Sanity check.
            if (instance.get() != nullptr)
            {
                // Do we have a sync worker thread?
                if (instance->sync_worker.get() != nullptr)
                {
                    // Close down the sync worker.
                    instance->sync_worker->close();
                }

                // Tell the server to shutdown.
                if (instance->grpc_server.get() != nullptr)
                {
                    instance->grpc_server->Shutdown(std::chrono::system_clock::now() + std::chrono::milliseconds(250));
                    instance->grpc_server.reset();
                }

                // Stop the thread.
                instance->stop();
                instance->join();

                // Reset the instance.
                instance.reset();
            }

            // We are now closed.
            isClosed = true;
        }
        TP(GRPC_FUNC, "%s: Done", __PRETTY_FUNCTION__);
    }

    void run() override;

    bool imageStreamReady();
    bool sendImage(const GRPC_IMAGE& image);
    bool sendNotification(uint32_t reg_address, uint32_t reg_value);

private:
    static const std::string DFLT_GRPC_ADDRESS;

    std::unique_ptr<GrpcSyncWorker> sync_worker;

    ServerBuilder builder;
    std::unique_ptr<Server> grpc_server;
};

//--------------------------------------------------------------------
// Static members are defined outside of the class.
//--------------------------------------------------------------------

//! Our local stream worker singleton.
std::unique_ptr<GrpcServer> GrpcServer::instance;
std::mutex GrpcServer::instance_mutex;

bool GrpcServer::isClosed(false);

// ======================================================
// ALM_TODO: Move the port value to the override object so that
//           it can be configured if need be.
// ======================================================
const std::string GrpcServer::DFLT_GRPC_ADDRESS("0.0.0.0:50051");

} /* End of GRPC_SERVER */

// Make things easier for us.
using namespace GRPC_SERVER;

//--------------------------------------------------------------------
// GrpcSyncWorker class implementation
//--------------------------------------------------------------------

GrpcSyncWorker::GrpcSyncWorker() :
    RDICameraService::Service(),
    closed(false),
    msg_id(0)
{
    // Create our requestor object.
    requestor = std::make_unique<GrpcRequestor>();
    if (requestor.get() != nullptr)
        requestor->start();
}

GrpcSyncWorker::~GrpcSyncWorker()
{
}

Status GrpcSyncWorker::AsyncNotify(ServerContext *context, ::grpc::ServerReaderWriter<AsyncNotifyResponse, NotifyEnable> *stream)
{
    std::unique_ptr<NOTIFY_INFO> stream_ptr;        // Our stream information structure.

    TP(GRPC_FUNC, "%s", __PRETTY_FUNCTION__);

    // Unused.
    (void)context;

    // Save our stream pointer.
    {
        std::unique_lock<std::mutex> locker(notify_mutex);

        // Build up our info structure.
        stream_ptr = std::make_unique<NOTIFY_INFO>();
        stream_ptr->enable = false;         // Writes will be handled in a separate function. Assume notifications are not enabled.
        stream_ptr->stream = stream;        // The writer we will use to stream notifications.

        // Put the stream information into our vector.
        notify_stream.emplace_back(stream_ptr.get());
    }

    // Perform the reads here. This will block and allow the GRPC server thread to continue processing.
    NotifyEnable request;
    while (stream->Read(&request) && !closed)
    {
        TP(GRPC_NOTIFY, "%s: request=%d", __PRETTY_FUNCTION__, request.enable());

        // Save the enable so the writer can do the correct thing.
        stream_ptr->enable = request.enable();
    }

    TP(GRPC_NOTIFY, "%s: Read() closed out.", __PRETTY_FUNCTION__);

    // Reset our steam pointer.
    if (!closed)
    {
        std::unique_lock<std::mutex> locker(notify_mutex);

        for (auto it = notify_stream.begin(); it != notify_stream.end(); ++it)
        {
            if (*it == stream_ptr.get())
            {
                notify_stream.erase(it);
                break;
            }
        }
    }

    // All done.
    TP(GRPC_NOTIFY, "%s: Done.", __PRETTY_FUNCTION__);
    return Status::OK;
}

Status GrpcSyncWorker::ImageStream(ServerContext *context, ::grpc::ServerReaderWriter<::RDICamServer::ImageData, ::RDICamServer::StreamEnable> *stream)
{
    std::unique_ptr<IMAGE_STREAM_INFO> stream_ptr;        // Our stream information structure.

    TP(GRPC_FUNC, "%s", __PRETTY_FUNCTION__);

    // Unused.
    (void)context;

    {
        std::unique_lock<std::mutex> locker(stream_mutex);

        // Build up our info structure.
        stream_ptr = std::make_unique<IMAGE_STREAM_INFO>();
        stream_ptr->enable = false;     // Writes will be handled in a separate function. We are good to go.
        stream_ptr->stream = stream;    // The writer we will use to stream.

        // Put the stream information into our vector.
        image_stream.emplace_back(stream_ptr.get());
    }

    // Perform the reads here. This will block and allow the GRPC server thread to continue processing.
    StreamEnable request;
    while (stream->Read(&request) && !closed)
    {
        TP(GRPC_STREAM, "%s: request=%d", __PRETTY_FUNCTION__, request.enable());

        // Save the enable so the writer can do the correct thing.
        stream_ptr->enable = request.enable();
    }

    TP(GRPC_STREAM, "%s: Read() closed out.", __PRETTY_FUNCTION__);

    // Reset our steam pointer.
    if (!closed)
    {
        std::unique_lock<std::mutex> locker(stream_mutex);

        for (auto it = image_stream.begin(); it != image_stream.end(); ++it)
        {
            if (*it == stream_ptr.get())
            {
                TP(GRPC_STREAM, "%s: Found it!", __PRETTY_FUNCTION__);
                image_stream.erase(it);
                break;
            }
        }
    }

    // All done.
    TP(GRPC_STREAM, "%s: Done.", __PRETTY_FUNCTION__);
    return Status::OK;
}

int GrpcSyncWorker::regReadCmdType(uint32_t reg_addr, uint16_t& reg_length)
{
    int rc(GIGE_MSG_READMEM_CMD);

    // Determine if the register address contains a string, uint32_t, byte array, etc.
    switch (reg_addr)
    {
    // ---------------------------
    // String, JSON, byte array, etc. values
    // ---------------------------

    // Use a default length when we don't know any better.
    case GIGE_REG_BOOTLOADER_VER:
    case GIGE_REG_COMMENT:
    case GIGE_REG_LUT_DATA:
    case GIGE_REG_PRNU_JSON:
    case GIGE_REG_SAVE_JSON:
    case GIGE_REG_FA_JSON:
        reg_length = GIGE_READMEM_B_LEN;
        break;
    case GIGE_REG_DEVICE_VER:
        reg_length = GIGE_REG_DEVICE_VER_B_LEN;
        break;
    case GIGE_REG_FILEPATH:
        reg_length = GIGE_FILEPATH_LEN;
        break;
    case GIGE_REG_IMAGE_TAG:
        reg_length = GIGE_IMAGE_TAG_MAX_LEN;
        break;
    case GIGE_REG_MAN_INFO:
        reg_length = GIGE_REG_MAN_INFO_B_LEN;
        break;
    case GIGE_REG_MAN_NAME:
        reg_length = GIGE_REG_MAN_NAME_B_LEN;
        break;
    case GIGE_REG_MODEL_NAME:
        reg_length = GIGE_REG_MODEL_NAME_B_LEN;
        break;
    case GIGE_REG_NTP_SERVER:
        reg_length = GIGE_NTP_SERVER_LEN;
        break;
    case GIGE_REG_SENSOR_VER:
        reg_length = GIGE_SENSOR_VER_LEN;
        break;
    case GIGE_REG_SERIAL_NUM:
        reg_length = GIGE_REG_SERIAL_NUM_B_LEN;
        break;
    case GIGE_REG_STREAM_FILEPATH:
        reg_length = GIGE_STREAM_FILEPATH_LEN;
        break;
    case GIGE_REG_TIMEZONE:
        reg_length = GIGE_TIMEZONE_LEN;
        break;
    case GIGE_REG_USER_DEF_NAME:
        reg_length = GIGE_REG_USER_DEF_NAME_B_LEN;
        break;
    case GIGE_REG_FFLR_JSON:
        reg_length = GIGE_FFLR_JSON_LEN;
        break;
    case GIGE_REG_LENS_INFO_JSON:
        reg_length = GIGE_LENS_INFO_JSON_LEN;
        break;
    case GIGE_REG_LENS_STATUS_JSON:
        reg_length = GIGE_LENS_STATUS_JSON_LEN;
        break;
    case GIGE_REG_LENS_SETTINGS_JSON:
        reg_length = GIGE_LENS_SETTINGS_JSON_LEN;
        break;
    case GIGE_REG_MEDIA_JSON:
        reg_length = GIGE_MEDIA_JSON_MAX_LEN;
        break;
    case GIGE_REG_PEEKPOKE_JSON:
        reg_length = GIGE_PEEKPOKE_JSON_LEN;
        break;
    case GIGE_REG_TELEMETRY_JSON:
        reg_length = GIGE_TELEMETRY_JSON_LEN;
        break;
    case GIGE_REG_FRAME_PFMD:
    case GIGE_REG_FRAME_METADATA:
        reg_length = GIGE_FRAME_METADATA_LEN;
        break;
    case GIGE_REG_VTRIG_SETTINGS_JSON:
        reg_length = GIGE_VTRIG_SETTINGS_JSON_LEN;
        break;

    default:
        // Most of the regisers are 32-bit unsigned integers.
        rc = GIGE_MSG_READREG_CMD;
        reg_length = GIGE_READREG_REQ_B_LEN;
        break;
    }

    return rc;
}

Status GrpcSyncWorker::ReadRegister(ServerContext *context, const ReadRegisterRequest *request, JsonResponse *response)
{
    uint16_t read_cmd;
    uint32_t reg_addr;

    // Ignore this for now.
    (void)context;

    // Are we shutting donw?
    if (closed)
    {
        fi_log(LOG_DEBUG, "%s: System in shutdown.", __PRETTY_FUNCTION__);
        return Status(StatusCode::INTERNAL, "System in shutdown.");
    }

    // Get the requested register address.
    reg_addr = request->reg_address();
    TP(GRPC_FUNC, "%s: addr=0x%04X", __PRETTY_FUNCTION__, reg_addr);

    // Build up our request to the server.
    GreqMessage cs_msg;
    buildReadRequest(reg_addr, read_cmd, cs_msg.msg);

    // Process the request.
    cs_msg.doSync(requestor.get());
    if (cs_msg.resp != FAS_THREAD::OK)
    {
        // Unexpected error.
        fi_log(LOG_DEBUG, "%s: The requestor returned a fault", __PRETTY_FUNCTION__);
        return Status(StatusCode::INTERNAL, "Unable to parse the requested register address!");
    }

    // Put the status into the response message.
    Json::Value read_resp;
    read_resp["status"] = ntohs(cs_msg.msg.message.header.resp.status);

    // Build up the register/value pairs for the response.
    Json::Value reg_pair = buildReadResponse(read_cmd, reg_addr, &cs_msg.msg.message);
    if (reg_pair.empty())
    {
        fi_log(LOG_DEBUG, "%s: buildReadResponse() returned an empty value.", __PRETTY_FUNCTION__);
        return Status(StatusCode::INTERNAL, "Unable to build a response!");
    }

    // Append the pair to our fields array.
    Json::Value reg_array;
    reg_array.append(reg_pair);
    read_resp["fields"] = reg_array;

    // Create a string from our JSON value.
    Json::StreamWriterBuilder wr_builder;
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";

    std::string json_message = Json::writeString(wr_builder, read_resp);
    TP(GRPC_READREG, "%s: Json:%s", __PRETTY_FUNCTION__, json_message.c_str());

    // Send it out the GRPC interface.
    response->mutable_reg_value()->assign(json_message);

    // All done.
    return Status::OK;
}

Status GrpcSyncWorker::ReadMultipleRegister(ServerContext *context, const MultipleRegisterRequest *request, JsonResponse *response)
{
    int cmd_index;
    int cmd_count;
    uint16_t status(0);
    uint32_t reg_addr;
    Json::Value reg_array;
    Json::Value read_resp;

    TP(GRPC_FUNC, "%s", __PRETTY_FUNCTION__);

    // Ignore this for now.
    (void)context;

    // Are we shutting donw?
    if (closed)
    {
        fi_log(LOG_DEBUG, "%s: System in shutdown.", __PRETTY_FUNCTION__);
        return Status(StatusCode::INTERNAL, "System in shutdown.");
    }

    // How many registers are we talking about.
    cmd_count = request->reg_addresses_size();
    TP(GRPC_READREG, "%s: cmd_count=%d", __PRETTY_FUNCTION__, cmd_count);
    for (cmd_index = 0; cmd_index < cmd_count; ++cmd_index)
    {
        uint16_t read_cmd;

        // Get the requested register address.
        reg_addr = request->reg_addresses(cmd_index);
        TP(GRPC_READREG, "%s: addr=0x%04X", __PRETTY_FUNCTION__, reg_addr);

        // Build up our request to the server.
        GreqMessage cs_msg;
        buildReadRequest(reg_addr, read_cmd, cs_msg.msg);

        // Process the request.
        cs_msg.doSync(requestor.get());
        if (cs_msg.resp != FAS_THREAD::OK)
        {
            // Unexpected error.
            fi_log(LOG_DEBUG, "%s: The requestor returned a fault", __PRETTY_FUNCTION__);
            return Status(StatusCode::INTERNAL, "Unable to parse the requested register address!");
        }

        // Did everything go OK?
        status = ntohs(cs_msg.msg.message.header.resp.status);
        TP(GRPC_READREG, "%s: addr=0x%04X, status=0x%x", __PRETTY_FUNCTION__, reg_addr, status);
        if (status != 0)
            break;

        // Build up the register/value pairs for the response.
        Json::Value reg_pair = buildReadResponse(read_cmd, reg_addr, &cs_msg.msg.message);
        if (reg_pair.empty())
        {
            fi_log(LOG_DEBUG, "%s: buildReadResponse() returned an empty value.", __PRETTY_FUNCTION__);
            return Status(StatusCode::INTERNAL, "Unable to build a response!");
        }

        // Append the pair to our fields array.
        reg_array.append(reg_pair);
    }

    // Put the status into the response message.
    read_resp["status"] = status;

    // Add our array.
    read_resp["fields"] = reg_array;

    // Create a string from our JSON value.
    Json::StreamWriterBuilder wr_builder;
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";

    std::string json_message = Json::writeString(wr_builder, read_resp);
    TP(GRPC_READREG, "%s: Json:%s", __PRETTY_FUNCTION__, json_message.c_str());

    // Send it out the GRPC interface.
    response->mutable_reg_value()->assign(json_message);

    // All done.
    return Status::OK;
}

Status GrpcSyncWorker::WriteRegisters(ServerContext *context, const WriteRegistersRequest *request, JsonResponse *response)
{
    uint16_t status(0);
    Json::Value json_request;
    Json::Reader json_reader;
    Json::Value resp_array;

    TP(GRPC_FUNC, "%s", __PRETTY_FUNCTION__);

    // Ignore this for now.
    (void)context;
    (void)response;

    // Are we shutting donw?
    if (closed)
    {
        fi_log(LOG_DEBUG, "%s: System in shutdown.", __PRETTY_FUNCTION__);
        return Status(StatusCode::INTERNAL, "System in shutdown.");
    }

    // Parse the request string into JSON.
    if (!json_reader.parse(request->write_string(), json_request))
    {
        fi_log(LOG_DEBUG, "%s: parse() failed!",  __PRETTY_FUNCTION__);
        return Status(StatusCode::INTERNAL, "parse() failed!");
    }

    // Make sure we have the "fields" key.
    if (json_request["fields"].isNull())
    {
        fi_log(LOG_DEBUG, "%s: Malformed request, no \"fields\" key!", __PRETTY_FUNCTION__);
        return Status(StatusCode::INTERNAL, "Malformed request, no \"fields\" key!");
    }

    // Get our array of values to write.
    Json::Value fields = json_request["fields"];

    // Walk through the array.
    GreqMessage cs_msg;
    for (Json::Value::ArrayIndex i = 0; i < fields.size(); ++i)
    {
        uint16_t host_sval;
        uint16_t msg_length;
        Json::Value entry = fields[i];

        // Each entry should be a pair of <addr>:<value>, where the addr is considered the "key".
        auto keys = entry.getMemberNames();
        std::string reg_addr_str = keys[0];
        uint32_t reg_addr = std::stoul(reg_addr_str, nullptr, 16);

        // Build up the read register message.
        memset(&cs_msg.msg, 0, sizeof(cs_msg.msg));
        cs_msg.msg.msg_type = CS_NETWORK::MSG_TYPE_GRPC;
        cs_msg.msg.message.header.req.code = GIGE_REQ_CODE;
        cs_msg.msg.message.header.req.flag = GIGE_FLAG_MUST_ACK_BE;

        // Are we looking at a string or an unsigned integer?
        Json::Value entry_value = entry[keys[0]];
        if (entry_value.isUInt())
        {
            uint32_t  reg_value = entry_value.asUInt();
            TP(GRPC_WRITEREG, "%s: reg_addr:0x%04X, reg_value=0x%08X", __PRETTY_FUNCTION__, reg_addr, reg_value);

            host_sval = GIGE_MSG_WRITEREG_CMD;
            cs_msg.msg.message.header.req.command = htons(host_sval);
            msg_length = GIGE_WRITEREG_REQ_B_LEN;
            cs_msg.msg.message.header.req.length = htons(msg_length);

            cs_msg.msg.message.data_l[0] = htonl(reg_addr);
            cs_msg.msg.message.data_l[1] = htonl(reg_value);
        }
        else
        {
            int buffer_len;
            std::string reg_value = entry_value.asString();
            TP(GRPC_WRITEREG, "%s: reg_addr:0x%04X, reg_value=%s", __PRETTY_FUNCTION__, reg_addr, reg_value);

            host_sval = GIGE_MSG_WRITEMEM_CMD;
            cs_msg.msg.message.header.req.command = htons(host_sval);

            // Figure out our buffer length. The GIGE_WRITEMEM_B_LEN value is already
            // adjusted for the size of the 'reg_addr' value.
            buffer_len = reg_value.size();
            if (buffer_len > GIGE_WRITEMEM_B_LEN)
                buffer_len = GIGE_WRITEMEM_B_LEN;

            msg_length = GIGE_WRITEMEM_ADDR_LEN + (uint16_t)buffer_len;
            cs_msg.msg.message.header.req.length = htons(msg_length);

            // Fill in our data.
            cs_msg.msg.message.data_l[0] = htonl(reg_addr);
            memcpy(&cs_msg.msg.message.data_l[1], reg_value.c_str(), (size_t)buffer_len);
        }

        msg_id++;                           // Increment before using.
        host_sval = msg_id;
        cs_msg.msg.message.header.req.req_id = htons(host_sval);

        // Process the request.
        cs_msg.doSync(requestor.get());
        if (cs_msg.resp != FAS_THREAD::OK)
        {
            // Unexpected error.
            fi_log(LOG_DEBUG, "%s: The requestor returned a fault", __PRETTY_FUNCTION__);
            return Status(StatusCode::INTERNAL, "Unable to parse the requested register address!");
        }


        // Did everything go OK?
        status = ntohs(cs_msg.msg.message.header.resp.status);

        // Save the register and write status.
        Json::Value resp_pair;
        resp_pair[reg_addr_str] = status;

        // Append it to our array.
        resp_array.append(resp_pair);

        // If we had an error, break out of the loop.
        if (status != 0)
        {
            fi_log(LOG_DEBUG, "%s: Bad status [0x%x] for write to reg: 0x%04X", __PRETTY_FUNCTION__, status, reg_addr);
            break;
        }
    }

    // Put the status into the response message.
    Json::Value write_resp;
    write_resp["status"] = status;

    // Add our array.
    write_resp["fields"] = resp_array;

    // Create a string from our JSON value.
    Json::StreamWriterBuilder wr_builder;
    wr_builder["commentStyle"] = "All";
    wr_builder["indentation"] = "";

    std::string json_message = Json::writeString(wr_builder, write_resp);
    TP(GRPC_WRITEREG, "%s: Json:%s", __PRETTY_FUNCTION__, json_message.c_str());

    // Send it out the GRPC interface.
    response->mutable_reg_value()->assign(json_message);

    return Status::OK;
}

void GrpcSyncWorker::close()
{
    TP(GRPC_FUNC, "%s", __PRETTY_FUNCTION__);

    // We are closed.
    closed = true;

    // Clean up our working threads.
    if (requestor.get() != nullptr)
    {
        // Tell the requestor to stop further processing.
        requestor->stop();

        // Is the thread still running?
        if (!requestor->isHalted())
        {
            FAS_THREAD::SyncMessage msg;

            // Tell the threads that we are stopping, if they haven't already stopped.
            msg.req = FAS_THREAD::HALT;

            // Pass it on. A no-op if already halted.
            msg.doSync(requestor.get());

            // Try to join the thread.
            requestor->join();
        }
    }

    TP(GRPC_FUNC, "%s: Done", __PRETTY_FUNCTION__);
}

bool GrpcSyncWorker::sendNotification(uint32_t reg_address, uint32_t reg_value)
{
    bool rc(false);

    // Are we shutting donw?
    if (!closed)
    {
        std::unique_lock<std::mutex> locker(notify_mutex);

        // Do we have anyone interested in a notification message?
        if (!notify_stream.empty())
        {
            TP(GRPC_FUNC, "%s: [0x%04x:0x%08x]", __PRETTY_FUNCTION__, reg_address, reg_value);

            // Build up our response.
            AsyncNotifyResponse response;
            response.set_reg_address(reg_address);
            response.set_reg_value(reg_value);

            // Loop through our vector of notify recipients.
            for (NOTIFY_INFO const *stream_info : notify_stream)
            {
                // Only do this if someone is listening.
                if (stream_info->enable)
                {
                    // Send the response.
                    rc = stream_info->stream->Write(response);
                }
            }
        }
    }

    // Return our result.
    return rc;
}

/*!
 * @brief The image stream is ready for images as long as
 *        we have a stream object to write to.
 *
 * @return True if we can stream images, false otherwise.
 */
bool GrpcSyncWorker::streamIsReady()
{
    // Are we shutting donw?
    if (closed)
        return false;

    return !image_stream.empty();
}

bool GrpcSyncWorker::streamSendImage(const GRPC_IMAGE& image)
{
    bool rc(false);

    // Are we shutting donw?
    if (!closed)
    {
        std::unique_lock<std::mutex> locker(stream_mutex);

        TP(GRPC_STREAM, "%s", __PRETTY_FUNCTION__);

        // Build up our response.
        ImageData response;
        response.set_format(image.format);
        response.set_width(image.width);
        response.set_height(image.height);
        response.set_frame_num(image.frame_num);

        // Do we need to append the image header?
        if ((image.header_len > 0) && (image.header_data != nullptr))
        {
            // Make sure the header data goes first, followed by the image data.
            response.mutable_image_data()->assign((const char *)image.header_data, image.header_len);
            response.mutable_image_data()->append((const char *)image.image_data, image.image_len);
        }
        else
        {
            // Just image data for this one.
            response.mutable_image_data()->assign((const char *)image.image_data, image.image_len);
        }

        // Loop through our vector of streams.
        for (IMAGE_STREAM_INFO const *stream_info : image_stream)
        {
            // Are we allowed to stream?
            if (stream_info->enable)
            {
                // Send the response.
                rc = stream_info->stream->Write(response);
            }
        }
    }

    // Return our result.
    return rc;
}

void GrpcSyncWorker::buildReadRequest(uint32_t reg_addr, uint16_t& read_cmd, CS_NETWORK::CS_NETWORK_MESSAGE& msg)
{
    uint16_t reg_length;

    // Clear the structure.
    memset(&msg, 0, sizeof(msg));

    // Initialize for a GRPC request.
    msg.msg_type = CS_NETWORK::MSG_TYPE_GRPC;
    msg.message.header.req.code = GIGE_REQ_CODE;
    msg.message.header.req.flag = GIGE_FLAG_MUST_ACK_BE;

    // Get and set the command.
    read_cmd = regReadCmdType(reg_addr, reg_length);
    msg.message.header.req.command = htons(read_cmd);
    if (read_cmd == GIGE_MSG_READREG_CMD)
    {
        // Set the request length for a register read.
        msg.message.header.req.length = htons(GIGE_READREG_REQ_B_LEN);

        // Fill in the data fields. Remember to convert to the expected network
        // byte order.
        msg.message.data_l[0] = htonl(reg_addr);
    }
    else
    {
        // Set the request length for a memory register read.
        uint16_t host_sval = GIGE_READMEM_REQ_B_LEN;
        msg.message.header.req.length = htons(host_sval);

        // Fill in the data fields. Remember to convert to the expected network
        // byte order.
        msg.message.data_l[GIGE_READMEM_REQ_ADDR_L] = htonl(reg_addr);

        // Sanity check the length.
        if (reg_length > GIGE_READMEM_B_LEN)
            reg_length = GIGE_READMEM_B_LEN;
        msg.message.data_w[GIGE_READMEM_REQ_COUNT_W] = htons(reg_length);
    }

    // Increment before using our msg_id.
    msg_id++;
    msg.message.header.req.req_id = htons(msg_id);

}

Json::Value GrpcSyncWorker::buildReadResponse(uint32_t read_cmd, uint32_t reg_addr, fcp_msg *message_ptr)
{
    Json::Value reg_pair;

    // We need the register address as a hex-string.
    std::string reg_string = fmt::format("0x{0:x}", reg_addr);

    // Build up the response. Remember to convert the values back to host byte order.
    if (read_cmd == GIGE_MSG_READREG_CMD)
    {
        // It was a straight read of a register value. Just move that over.
        reg_pair[reg_string] = ntohl(message_ptr->data_l[0]);
    }
    else
    {
        uint16_t data_length = ntohs(message_ptr->header.resp.length);
        if ((data_length <= GIGE_READMEM_ACK_DATA_B) || (data_length > FCP_MSG_DATA_SIZE))
        {
            fi_log(LOG_DEBUG, "%s: Error reading reg_addr:0x%04x, data_length:%d", __PRETTY_FUNCTION__, reg_addr, data_length);
            reg_pair[reg_string] = std::string(" ");
        }
        else
        {
            std::string data_string;

            // The acutal data starts at offset GIGE_READMEM_ACK_DATA_B, which skips over the
            // register address being read. Adjust things appropriately.
            data_length -= GIGE_READMEM_ACK_DATA_B;
            data_string = std::string((const char *)&message_ptr->data_b[GIGE_READMEM_ACK_DATA_B], data_length);

            // If the register being read is truly a string register, make sure we trim out all the fat.
            // Determine if the register address contains a string, uint32_t, byte array, etc.
            switch (reg_addr)
            {
            // ---------------------------
            // String values
            // ---------------------------
            case GIGE_REG_BOOTLOADER_VER:
            case GIGE_REG_COMMENT:
            case GIGE_REG_DEVICE_VER:
            case GIGE_REG_FILEPATH:
            case GIGE_REG_IMAGE_TAG:
            case GIGE_REG_MAN_INFO:
            case GIGE_REG_MAN_NAME:
            case GIGE_REG_MODEL_NAME:
            case GIGE_REG_NTP_SERVER:
            case GIGE_REG_SENSOR_VER:
            case GIGE_REG_SERIAL_NUM:
            case GIGE_REG_STREAM_FILEPATH:
            case GIGE_REG_TIMEZONE:
            case GIGE_REG_USER_DEF_NAME:

            // ---------------------------
            // Json values
            // ---------------------------
            case GIGE_REG_FFLR_JSON:
            case GIGE_REG_LENS_INFO_JSON:
            case GIGE_REG_LENS_STATUS_JSON:
            case GIGE_REG_LENS_SETTINGS_JSON:
            case GIGE_REG_MEDIA_JSON:
            case GIGE_REG_PEEKPOKE_JSON:
            case GIGE_REG_TELEMETRY_JSON:
            case GIGE_REG_VTRIG_SETTINGS_JSON:
            case GIGE_REG_PRNU_JSON:
            case GIGE_REG_SAVE_JSON:
            case GIGE_REG_FA_JSON:
                // Trim trailing nulls out of the string.
                data_string.erase(std::find(data_string.begin(), data_string.end(), '\0'), data_string.end());
                break;

            // ---------------------------
            // Byte array values
            // ---------------------------
            case GIGE_REG_FRAME_PFMD:
            case GIGE_REG_FRAME_METADATA:
            case GIGE_REG_LUT_DATA:
                // We need to base64 encode these values.
                data_string = macaron::Base64::Encode(data_string);
                break;

            default:
                break;
            }

            // Save the value into the pair.
            reg_pair[reg_string] = data_string;
        }
    }

    // Return our JSON encoded pair.
    return reg_pair;
}

//--------------------------------------------------------------------
// GrpcServer class implementation
//--------------------------------------------------------------------

/*!
 * @brief The worker thread constructor. The worker thread
 *    is a "detached" thread. That is, it runs to completion (once
 *    started) and does not expect to "join" with any other thread.
 */
GrpcServer::GrpcServer() :
    FAS_THREAD::Thread(true)
{
    LOG("%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Cleanup the thread when it's been told to exit.
 */
GrpcServer::~GrpcServer()
{
}

GrpcServer *GrpcServer::factory()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    // Check to see if we have our singleton and we have not closed.
    if ((instance.get() == nullptr) && !isClosed)
    {
        // Create the one and only instance.
        instance = std::unique_ptr<GrpcServer>(new GrpcServer());
    }

    return instance.get();
}

void GrpcServer::run()
{
    std::string grpc_address = GrpcServer::DFLT_GRPC_ADDRESS;

    LOG("%s: PID=%u", __PRETTY_FUNCTION__, get_tid());

    // Allocate our worker objects.
    sync_worker = std::unique_ptr<GrpcSyncWorker>(new GrpcSyncWorker);
    if (sync_worker.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate the GrpcSyncWorker object!");
        return;
    }

    // Tell the builder the URL we'll be listening to.
    builder.AddListeningPort(grpc_address, grpc::InsecureServerCredentials());

    // Tell the builder what object is going to be doing the work. Add the synchronous
    // server.
    builder.RegisterService(sync_worker.get());

    // Have the builder start up the server. It will return a unique_ptr that
    // we can use.
    grpc_server = builder.BuildAndStart();
    if (grpc_server.get() == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to allocate the Server object!", FI_FUNCTION);
        return;
    }
    fi_log(LOG_DEBUG, "%s: GRPC Server listening on %s", FI_FUNCTION, grpc_address.c_str());

    // Block here until we are shutting down.
    //
    // ----------------------------------------------------------------------
    // NOTE: This wait call is optional. At this point in time, I just don't
    //       know what other trouble I can get up to.
    // ----------------------------------------------------------------------
    grpc_server->Wait();
}

bool GrpcServer::imageStreamReady()
{
    if (sync_worker.get() == nullptr)
        return false;
    return sync_worker->streamIsReady();
}

bool GrpcServer::sendImage(const GRPC_IMAGE& image)
{
    if (sync_worker.get() == nullptr)
        return false;
    return sync_worker->streamSendImage(image);
}

bool GrpcServer::sendNotification(uint32_t reg_address, uint32_t reg_value)
{
    if (sync_worker.get() == nullptr)
        return false;
    return sync_worker->sendNotification(reg_address, reg_value);
}

//--------------------------------------------------------------------
// Access functions
//--------------------------------------------------------------------

/*!
 * @brief Initialize and start the GRPC server thread.
 *
 * @returns non-zero for failure, 0 for success.
 */
int grpcServerInit()
{
    TP(GRPC_FUNC, "%s", __PRETTY_FUNCTION__);

    GrpcServer *grpcPtr = GrpcServer::factory();
    if (grpcPtr != nullptr)
        grpcPtr->start();
    return 0;
}

/*!
 * @brief Close down the GRPC server thread.
 * @return non-zero for failure, 0 for success.
 */
int grpcServerClose()
{
    TP(GRPC_FUNC, "%s", __PRETTY_FUNCTION__);

    GrpcServer::close();
    return 0;
}

/*!
 * @brief Send a response back to the client, for a message request we have already received.
 *
 * @param p_msg     Message which is destined for the client.
 * @param num_bytes Length of the message to transmit.
 *
 * @return the size of the message if successful, 0 if no message, <0 if error
 */
int grpcSendMsg(CS_NETWORK::CS_NETWORK_MESSAGE *p_msg, std::uint32_t num_bytes)
{
    TP(GRPC_REQUESTOR, "%s: req_id=%d, num_bytes=%d/%d", __PRETTY_FUNCTION__, htons(p_msg->message.header.req.req_id), num_bytes, sizeof(CS_NETWORK::CS_NETWORK_MESSAGE));

    // Pass it on to the requestor.
    GrpcRequestor::receiveResponse(p_msg);
    return num_bytes;
}

/*!
 * @brief Pass a register address/value pair on to be sent as an asynchronous notification.
 *
 * @param reg_address   The address of the register.
 * @param reg_value     The value of the register.
 *
 * @returns True if successful, false otherwise.
 */
bool grpcNotification(uint32_t reg_address, uint32_t reg_value)
{
    bool rc(false);

    GrpcServer *grpcPtr = GrpcServer::factory();
    if (grpcPtr != nullptr)
        rc = grpcPtr->sendNotification(reg_address, reg_value);

    return rc;
}

/*!
 * @brief Checks to see if an image stream has been opened.
 * @return True if we are ready to stream, false otherwise.
 */
bool grpcStreamIsConnected()
{
    bool rc(false);

    GrpcServer *grpcPtr = GrpcServer::factory();
    if (grpcPtr != nullptr)
        rc = grpcPtr->imageStreamReady();

    return rc;
}

/*!
 * @brief Write an image out to the image stream.
 * @return True if we are successful, false otherwise.
 */
bool grpcStreamImage(const GRPC_IMAGE& image)
{
    bool rc(false);

    GrpcServer *grpcPtr = GrpcServer::factory();
    if (grpcPtr != nullptr)
        rc = grpcPtr->sendImage(image);

    return rc;
}
